# This file is part of tridiax, a toolkit for solving tridiagonal systems. tridiax is
# licensed under the Apache License Version 2.0, see <https://www.apache.org/licenses/>

VERSION = (0, 2, 0)

__version__ = ".".join(map(str, VERSION))
